{-# LANGUAGE BangPatterns        #-}
{-# LANGUAGE LambdaCase          #-}
{-# LANGUAGE OverloadedStrings   #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TemplateHaskell     #-}
{-# LANGUAGE TypeApplications    #-}
{-# LANGUAGE ViewPatterns        #-}

module Gargantext.API.Server.Named.Remote (
  remoteAPI
  ) where

import Codec.Serialise
import Conduit
import Control.Exception.Safe qualified as Safe
import Control.Exception (toException)
import Control.Lens (view, (#))
import Control.Monad.Except (throwError, MonadError)
import Control.Monad (void, liftM2, forM_)
import Data.Aeson qualified as JSON
import Data.Aeson.Types qualified as JS
import Data.ByteString.Builder qualified as B
import Data.ByteString.Lazy qualified as BL
import Data.Conduit.Combinators qualified as C
import Data.Conduit.List qualified as CL
import Data.Foldable (for_, foldlM)
import Data.List.Split qualified as Split
import Data.Text.Encoding  qualified as TE
import Data.Text qualified as T
import Gargantext.API.Admin.Auth
import Gargantext.API.Admin.Auth.Types (AuthenticatedUser(..))
import Gargantext.API.Auth.PolicyCheck (remoteExportChecks)
import Gargantext.API.Errors.Types
import Gargantext.API.Ngrams.Prelude (getNgramsList)
import Gargantext.API.Ngrams.Types (NgramsList)
import Gargantext.API.Node.Document.Export (get_document_json)
import Gargantext.API.Node.Document.Export.Types
import Gargantext.API.Prelude (IsGargServer)
import Gargantext.API.Routes.Client (remoteImportClient)
import Gargantext.API.Routes.Named.Remote qualified as Named
import Gargantext.Core.Config
import Gargantext.Core (lookupDBid)
import Gargantext.Core.NLP (HasNLPServer)
import Gargantext.Core.NodeStory.Types (HasNodeStoryEnv, HasNodeArchiveStoryImmediateSaver)
import Gargantext.Core.Types.Main
import Gargantext.Core.Worker.Jobs (sendJob)
import Gargantext.Core.Worker.Jobs.Types qualified as Jobs
import Gargantext.Database.Admin.Types.Hyperdata.Frame (HyperdataFrame(..))
import Gargantext.Database.Admin.Types.Node hiding (WARNING, INFO)
import Gargantext.Database.Prelude (IsDBCmd)
import Gargantext.Database.Query.Table.Node.Error (HasNodeError, nodeError, NodeError (..))
import Gargantext.Database.Query.Table.Node (insertNodeWithHyperdata, getNodes, getUserRootPrivateNode)
import Gargantext.Database.Schema.Node (NodePoly(..))
import Gargantext.Orphans ()
import Gargantext.System.Logging (logLocM, LogLevel(..), MonadLogger)
import GHC.Generics (Generic)
import Network.HTTP.Client qualified as HTTP
import Prelude
import Servant.Client.Streaming (mkClientEnv, withClientM, ClientError)
import Servant.Server.Generic (AsServerT)

data ExportableNode = ExportableNode
  { _en_node         :: Node JSON.Value
  , _en_node_payload :: Maybe ExportableNodePayload
  } deriving Generic

data ExportableNodePayload
  = -- | If this node is a \"docs\" node, remotely export also
    -- all the associated documents.
    ENP_document   DocumentExport
    -- | If this node is a \"terms\" node, remotely export also
    -- all the associated ngrams
  | ENP_terms      NgramsList
    -- | If this node is a \"note\" node, remotely export also
    -- all the raw markdown blob
  | ENP_notes T.Text
  deriving Generic

instance Serialise ExportableNodePayload where
instance Serialise ExportableNode where

remoteAPI :: (MonadIO m, IsGargServer env BackendInternalError m, HasNodeArchiveStoryImmediateSaver env)
          => AuthenticatedUser
          -> Named.RemoteAPI (AsServerT m)
remoteAPI authenticatedUser = Named.RemoteAPI $
  Named.RemoteAPI'
    { remoteExportEp = \payload@Named.RemoteExportRequest{..} mgr ->
        withPolicy authenticatedUser (remoteExportChecks _rer_node_id) (remoteExportHandler authenticatedUser payload) mgr
    , remoteImportEp = remoteImportHandler authenticatedUser
    }

type ExpectedPayload = Tree ExportableNode

remoteImportHandler :: forall err env m.
                    ( HasNodeStoryEnv env
                    , HasNodeError err
                    , HasBackendInternalError err
                    , HasNodeArchiveStoryImmediateSaver env
                    , IsDBCmd env err m
                    , HasNLPServer env
                    , MonadLogger m
                    , MonadIO m)
                    => AuthenticatedUser
                    -> ConduitT () Named.RemoteBinaryData IO ()
                    -> m [NodeId]
remoteImportHandler loggedInUser c = do
  chunks <- liftIO $ sourceToList $ c .| C.map (B.byteString . Named.getRemoteBinaryData)
  -- FIXME(adn): We have to find a way to deserialise this into a streaming fashion and
  -- attempt insertion one element of the list at the time.
  case deserialiseOrFail @ExpectedPayload (B.toLazyByteString $ mconcat chunks) of
    Left err -> throwError $ _BackendInternalError # InternalUnexpectedError (toException $ userError $ "Deserialization error: " ++ show err)
    Right (TreeN x xs) -> do
      $(logLocM) INFO $ "Importing " <> T.pack (show $ _node_id $ _en_node $ x)
      -- NOTE(adn) By default, we append the imported node(s) to the user's
      -- private folder.
      privateFolderId <- _node_id <$> getUserRootPrivateNode (_auth_user_id loggedInUser)
      $(logLocM) INFO $ "Attaching " <> T.pack (show $ _node_id $ _en_node $ x) <> " to private folder " <> T.pack (show privateFolderId)
      -- Attempts to insert nodes a we go along.
      rootNode <- insertNode (Just privateFolderId) x
      nodes <- foldlM (insertTrees (Just rootNode)) [rootNode] xs
      $(logLocM) INFO $ "Successfully imported all the requested nodes."
      pure nodes
  where

    insertNode :: Maybe NodeId -> ExportableNode -> m NodeId
    insertNode mb_parent (ExportableNode x mb_payload) = case lookupDBid $ _node_typename x of
      Nothing -> throwError $ _BackendInternalError # InternalUnexpectedError (toException $ userError $ "remoteImportHandler: impossible, node with invalid type.")
      Just ty -> do
        new_node <- insertNodeWithHyperdata ty (_node_name x) (_node_hyperdata x) mb_parent (_auth_user_id loggedInUser)
        $(logLocM) INFO $ "Created a new node " <> T.pack (show $ new_node) <> " of type " <> T.pack (show ty)
        for_ (liftM2 (,) mb_payload mb_parent) $ \(exported_payload, parentId) -> do
          case exported_payload of
            ENP_document   docsList
              -> do
                $(logLocM) INFO $ "Found document list to import..."
                let totalDocs = _de_documents docsList
                let chunks = Split.chunksOf 100 totalDocs
                forM_ (zip [1..] chunks) $ \(local_ix, chunk) -> do
                  let ws = Jobs.WorkSplit
                        { Jobs._ws_current = min (length totalDocs) (((local_ix - 1) * length chunk) + length chunk)
                        , Jobs._ws_total   = length totalDocs
                        }
                  let payload = Jobs.ImportRemoteDocumentsPayload loggedInUser parentId new_node chunk ws
                  void $ sendJob $ Jobs.ImportRemoteDocuments payload
            ENP_terms ngramsList
              -> do
                $(logLocM) INFO $ "Found ngrams list to import..."
                void $ sendJob $ Jobs.ImportRemoteTerms $ Jobs.ImportRemoteTermsPayload new_node ngramsList
            ENP_notes _noteAsMarkdown
              -> do
                $(logLocM) INFO $ "Found some markdown notes to import"
                -- FIXME(adn) actually import the notes

        pure new_node

    insertTrees :: Maybe NodeId -> [NodeId] -> Tree ExportableNode -> m [NodeId]
    insertTrees currentParent !acc (TreeN x xs) = do
      childrenRoot <- insertNode currentParent x
      (`mappend` acc) <$> foldlM (insertTrees (Just childrenRoot)) [childrenRoot] xs

remoteExportHandler :: ( MonadIO m, Safe.MonadCatch m
                       , IsGargServer err env m
                       )
                    => AuthenticatedUser
                    -> Named.RemoteExportRequest
                    -> m [NodeId]
remoteExportHandler loggedInUser Named.RemoteExportRequest{..} = do
  mgr  <- view gargHttpManager
  nodes <- getNodes _rer_node_id
  checkNodesTypeAllowed nodes
  exportable <- makeExportable (_auth_node_id loggedInUser) nodes
  liftIO (withClientM (remoteImportClient _rer_instance_auth (streamEncoder exportable)) (mkClientEnv mgr _rer_instance_url) streamDecode)
              `Safe.catch` \(e :: BackendInternalError) -> throwError $ _BackendInternalError # e

makeExportable :: (MonadIO m, IsGargServer err env m)
               => NodeId
               -> Tree (Node JSON.Value)
               -> m (Tree ExportableNode)
makeExportable userNodeId (TreeN x xs)
  | Just nty <- lookupDBid (_node_typename x)
  = do
    mb_payload <- case nty of
      NodeTexts -> Just . ENP_document <$> get_document_json userNodeId (_node_id x)
      NodeList  -> Just . ENP_terms    <$> getNgramsList (_node_id x)
      Notes     -> case JS.parseMaybe JS.parseJSON (_node_hyperdata x) of
        Nothing
          -> pure Nothing
        Just hframe
          -> do
            mgr  <- view gargHttpManager
            tryExportNode mgr hframe
      _         -> pure Nothing
    let exportableRoot = ExportableNode x mb_payload
    children <- mapM (makeExportable userNodeId) xs
    pure $ TreeN exportableRoot children
  | otherwise
  = throwError $ _BackendInternalError
               # InternalUnexpectedError (toException $ userError $ "remoteImportHandler: impossible, node with invalid type.")

tryExportNode :: (MonadIO m, MonadLogger m) => HTTP.Manager -> HyperdataFrame -> m (Maybe ExportableNodePayload)
tryExportNode mgr HyperdataFrame{..} = do
  let download_url = _hf_base <> "/" <> _hf_frame_id <> "/download"
  case HTTP.parseRequest (T.unpack download_url) of
    Left err -> do
      $(logLocM) WARNING $ "Couldn't extract a valid URL from " <> download_url <> ", " <> T.pack (show err)
      pure Nothing
    Right rq -> do
      res <- HTTP.responseBody <$> liftIO (HTTP.httpLbs rq mgr)
      pure $ Just $ ENP_notes (TE.decodeUtf8 $ BL.toStrict $ res)

checkNodesTypeAllowed :: (MonadError e m, HasNodeError e) => Tree (Node a) -> m ()
checkNodesTypeAllowed (TreeN r xs) = do
  checkNodeTypeAllowed r
  mapM_ checkNodesTypeAllowed xs

checkNodeTypeAllowed :: (MonadError e m, HasNodeError e) => Node a -> m ()
checkNodeTypeAllowed n
  | Just nty <- lookupDBid (_node_typename n)
  , nty `elem` exportableNodeTypes
  = pure ()
  | otherwise
  = let msg = "It's possible to export only the following node of type: " <> T.intercalate "," (map (T.pack . show) exportableNodeTypes)
    in nodeError $ NodeNotExportable (_node_id n) msg

-- | At the moment we support only export corpus nodes and their children (i.e. "Docs", "Terms", "Graph").
exportableNodeTypes :: [NodeType]
exportableNodeTypes = [ NodeCorpus, NodeCorpusV3, NodeTexts, NodeGraph, NodeList, NodePhylo, Notes ]

streamEncoder :: MonadIO m => ExpectedPayload -> ConduitT () Named.RemoteBinaryData m ()
streamEncoder = CL.sourceList . map Named.RemoteBinaryData . BL.toChunks . serialise

-- | Returns a conduit which can be used to decode
streamDecode :: Either ClientError [NodeId] -> IO [NodeId]
streamDecode = \case
  Left err -> Safe.throwIO $ InternalUnexpectedError (toException $ userError $ show err)
  Right x  -> pure x
